#!/usr/bin/env python
"""
MIRAR: https://github.com/hardbyte/python-can/blob/main/examples/asyncio_demo.py
pi@raspberrypi:~/codis/canbus $ sudo ip link set can0 type can bitrate 100000
pi@raspberrypi:~/codis/canbus $ sudo ifconfig can0 up
pi@raspberrypi:~/codis/canbus $ sudo ifconfig can0 down

"""
import os
import can
from datetime import datetime
from time import sleep
import RPi.GPIO as GPIO

USE_LEDS = True
VERBOSE = False
LED_W = 12
LED_B = 13
TIMEOUT_RCV = 0.5

if USE_LEDS:
    GPIO.setmode(GPIO.BCM)
    GPIO.setwarnings(False)
    GPIO.setup(LED_W,GPIO.OUT)
    GPIO.setup(LED_B,GPIO.OUT)

def doutMsg(addr,side,value) -> can.Message:
    global VERBOSE
    i2cAddr = int(addr,2)
    msg = []
    msg.append(2 if side.lower() == 'a' else 18)
    msg.append(value)
    if VERBOSE:
        print(msg)
    return can.Message(
        arbitration_id = 800 + i2cAddr,
        data = msg,
        is_extended_id = False,
        is_remote_frame = False,
        dlc = 2,
        timestamp = datetime.timestamp(datetime.now())
    )

def dversionMsg(addr) -> can.Message:
    i2cAddr = int(addr,2)
    msg = []
    return can.Message(
        arbitration_id = 800 + i2cAddr,
        data = msg,
        is_extended_id = False,
        is_remote_frame = True,
        dlc = 0,
        timestamp = datetime.timestamp(datetime.now())
    )

def send_can(msg):
    global USE_LEDS
    if USE_LEDS:
        GPIO.output(LED_W,GPIO.HIGH)
    bus.send(msg)
    if USE_LEDS:
        GPIO.output(LED_W,GPIO.LOW)

def recv_can():
    global USE_LEDS
    if USE_LEDS:
        GPIO.output(LED_B,GPIO.HIGH)
    msg = bus.recv(TIMEOUT_RCV)
    if USE_LEDS:
        GPIO.output(LED_B,GPIO.LOW)
    return msg
   
def dout(addr,side,value):
    send_can(doutMsg(addr,side,value))

def dversion(addr):
    global VERBOSE
    send_can(dversionMsg(addr))
    msg = bus.recv(.5)
    if VERBOSE:
        print(msg)
    if msg is not None:
        version = "%d.%d"%(msg.data[1],msg.data[2])
        if VERBOSE:
            # print(msg.data)
            print("Rib cfg: %d, version: %d.%d"%(msg.data[0],msg.data[1],msg.data[2]))
    else:
        version = "0.0"
        if VERBOSE:
            print("No CAN answer in %.2f seconds"%TIMEOUT_RCV)
    return version

if __name__ == "__main__":
    # # os.system('sudo ip link set can0 type can bitrate 100000')
    # # os.system('sudo ifconfig can0 up') # Enable can0 
    # os.system('sudo ip link set up can0 type can bitrate 100000')
    try:
        bus = can.interface.Bus(channel = 'can0', bustype = 'socketcan')

        print("version: %s"%dversion('0000'))
        sleep(0.5)

        dout('0000','A',0xFF)
        print("0xFF")
        sleep(0.5)
        dout('0000','A',0xA5)
        print("0xA5")
        sleep(0.5)
        dout('0000','A',0x5A)
        print("0x5A")
        sleep(0.5)
        dout('0000','A',0x00)
        print("0x00")
        sleep(0.5)
    except KeyboardInterrupt:
        dout('0000','A',0x00)
        print("0x00")
        sleep(0.1)
        # os.system('sudo ifconfig can0 down')  #Disable can0

